"""
Handles reading and writing of Equilibrium objects

Writing is relatively straightforward, but reading requires inferring
the currents in the PF coils

Copyright 2016 Ben Dudson, University of York. Email: benjamin.dudson@york.ac.uk

This file is part of FreeGS.

FreeGS is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

FreeGS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with FreeGS.  If not, see <http://www.gnu.org/licenses/>.

"""

from . import _geqdsk
from . import critical
from .equilibrium import Equilibrium
from . import jtor
from . import control
from . import picard

from scipy import interpolate
from numpy import linspace, amin, amax, reshape, ravel, zeros, argmax, clip, sin, cos, pi

def write(eq, fh, label=None, oxpoints=None, fileformat=_geqdsk.write):
    """
    Write a GEQDSK equilibrium file, given a FreeGS Equilibrium object
    
    eq - Equilibrium object
    fh - file handle
    
    label - Text label to put in the file
    oxpoints - O- and X-points  (opoint, xpoint) returned by critical.find_critical
          If not given, it will be calculated using critical.find_critical
    """
    # Get poloidal flux
    psi = eq.psi()

    # Get size of the grid
    nx,ny = psi.shape

    if oxpoints:
        opoint, xpoint = oxpoints
    else:
        # Find the O- and X-points
        opoint, xpoint = critical.find_critical(eq.R, eq.Z, psi)

    rmin = eq.Rmin
    rmax = eq.Rmax
    zmin = eq.Zmin
    zmax = eq.Zmax

    fvac = eq.fvac() # Vacuum f = R*Bt
    R0 = 1.0 # Reference location
    B0 = fvac / R0 # Reference vacuum toroidal magnetic field
    
    data = {"nx":nx, "ny":ny,
            "rdim":rmax-rmin, # Horizontal dimension in meter of computational box
            "zdim":zmax-zmin, # Vertical dimension in meter of computational box
            "rcentr":R0, # R in meter of vacuum toroidal magnetic field BCENTR
            "bcentr":fvac/R0, # Vacuum magnetic field at rcentr
            "rleft":rmin, # Minimum R in meter of rectangular computational box
            "zmid":0.5*(zmin + zmax)} # Z of center of computational box in meter
            
    
    data["rmagx"], data["zmagx"], data["simagx"] = opoint[0] # magnetic axis
    
    data["sibdry"] = xpoint[0][2]  # Psi at boundary
    
    data["cpasma"] = eq.plasmaCurrent() # Plasma current [A]

    psinorm = linspace(0.0, 1.0, nx, endpoint=False) # Does not include separatrix

    data["fpol"] = eq.fpol(psinorm)
    data["pres"] = eq.pressure(psinorm)

    data["psi"] = psi
    
    qpsi = zeros([nx])
    qpsi[1:] = eq.q(psinorm[1:]) # Exclude axis
    qpsi[0] = qpsi[1]
    data["qpsi"] = qpsi
    
    if hasattr(eq, "rlim") and hasattr(eq, "zlim"):
        data["rlim"] = eq.rlim
        data["zlim"] = eq.zlim
    
    # Call fileformat to write the data
    fileformat(data, fh, label=label)

import matplotlib.pyplot as plt

def read(fh, machine, rtol=1e-3, ntheta=8, show=False, axis=None):
    """
    Reads a G-EQDSK format file
    
    fh      - file handle
    machine - a Machine object defining coil locations
    rtol    - Relative error in nonlinear solve
    ntheta  - Number of points in poloidal angle on the separatrix
              this is used to constrain the plasma shape
    show    - Set to true to show solution in a new figure
    axis    - Set to an axis for plotting. Implies show=True
    
    A nonlinear solve will be performed, using Picard iteration
    
    Returns
    -------

    An Equilibrium object eq. In addition, the following is available:

    eq.control   - The control system
    eq._profiles - The profiles object
    
    """

    if axis is not None:
        show = True

    # Read the data as a Dictionary
    data = _geqdsk.read(fh)

    # Create an Equilibrium object
    eq = Equilibrium(tokamak = machine,
                     Rmin = data["rleft"],
                     Rmax = data["rleft"] + data["rdim"],
                     Zmin = data["zmid"] - 0.5*data["zdim"],
                     Zmax = data["zmid"] + 0.5*data["zdim"],
                     nx=data["nx"], ny=data["ny"]         # Number of grid points
                 )

    # Range of psi normalises psi derivatives
    psirange = data["sibdry"] - data["simagx"]
    
    psinorm = linspace(0.0, 1.0, data["nx"], endpoint=True)
    
    # Create a spline fit to pressure, f and f**2
    p_spl = interpolate.InterpolatedUnivariateSpline(psinorm, data["pres"])
    pprime_spl = interpolate.InterpolatedUnivariateSpline(psinorm, data["pres"] / psirange).derivative()
    
    f_spl = interpolate.InterpolatedUnivariateSpline(psinorm, data["fpol"])
    ffprime_spl = interpolate.InterpolatedUnivariateSpline(psinorm, 0.5*data["fpol"]**2/psirange).derivative() 

    # functions to return p, pprime, f and ffprime
    def p_func(psinorm):
        if hasattr(psinorm, "shape"):
            return reshape(p_spl(ravel(psinorm)),psinorm.shape)
        return p_spl(psinorm)

    def f_func(psinorm):
        if hasattr(psinorm, "shape"):
            return reshape(f_spl(ravel(psinorm)),psinorm.shape)
        return f_spl(psinorm)

    def pprime_func(psinorm):
        if hasattr(psinorm, "shape"):
            return reshape(pprime_spl(ravel(psinorm)),psinorm.shape)
        return pprime_spl(psinorm)
    
    def ffprime_func(psinorm):
        if hasattr(psinorm, "shape"):
            return reshape(ffprime_spl(ravel(psinorm)),psinorm.shape)
        return ffprime_spl(psinorm)

    
    # Create a set of profiles to calculate toroidal current density Jtor
    profiles = jtor.ProfilesPprimeFfprime(pprime_func,
                                          ffprime_func,
                                          data["rcentr"] * data["bcentr"],
                                          p_func=p_func, 
                                          f_func=f_func)

    # Use these profiles to calculate plasma psi
    # This requires a bit of a hack to set the poloidal flux
    
    coil_psi = machine.psi(eq.R, eq.Z)
    eq._updatePlasmaPsi(data["psi"] - coil_psi)
    
    # Perform a linear solve, calculating plasma psi
    eq.solve(profiles)

    print("Plasma current: {0} Amps, input: {1} Amps".format(eq.plasmaCurrent(), data["cpasma"]))
    
    # Identify points to constrain: X-points, O-points and separatrix shape

    psi_in = interpolate.RectBivariateSpline(eq.R[:,0], eq.Z[0,:], data["psi"])
    
    # Find all the O- and X-points
    opoint, xpoint = critical.find_critical(eq.R, eq.Z, data["psi"])
    
    if show:
        if axis is None:
            fig = plt.figure()
            axis = fig.add_subplot(111)
        axis.set_aspect('equal')
        axis.set_xlabel("Major radius [m]")
        axis.set_ylabel("Height [m]")
    
        levels = linspace(amin(data["psi"]), amax(data["psi"]), 50)
        axis.contour(eq.R,eq.Z,data["psi"], levels=levels, colors='k')
        
        # Put red dots on X-points
        for r,z,_ in xpoint:
            axis.plot(r,z,'ro')

        sep_contour=axis.contour(eq.R, eq.Z, data["psi"], levels=[xpoint[0][2]], colors='r')
        
    # Find the separatrix
    isoflux = critical.find_separatrix(eq, opoint, xpoint, ntheta=ntheta, psi=data["psi"], axis=axis)

    # Find best fit for coil currents
    controlsystem = control.constrain(xpoints=xpoint, isoflux=isoflux, gamma=1e-14)
    controlsystem(eq)
    
    if show:
        axis.contour(eq.R,eq.Z, eq.psi(), levels=levels, colors='r')
        plt.pause(1)
    
    # Print the coil currents
    machine.printCurrents()

    ####################################################################
    # Refine the equilibrium to ensure consistency 
    # Solve using Picard iteration
    #
    
    picard.solve(eq,          # The equilibrium to adjust
                 profiles,    # The toroidal current profile function
                 controlsystem, show=show, axis=axis,
                 rtol=rtol)
    
    print("Plasma current: {0} Amps, input: {1} Amps".format(eq.plasmaCurrent(), data["cpasma"]))
    print("Plasma pressure on axis: {0} Pascals".format(eq.pressure(0.0)))
    machine.printCurrents()

    # Save the control system to eq
    eq.control = controlsystem
    
    return eq



"""
 Classes representing plasma profiles.

 These must have the following methods:

   Jtor(R, Z, psi) 
      -> Return a numpy array of toroidal current density [J/m^2]
   pprime(psinorm)
      -> return p' at given normalised psi
   ffprime(psinorm)
      -> return ff' at given normalised psi
   pressure(psinorm)
      -> return p at given normalised psi
   fpol(psinorm)
      -> return f at given normalised psi
   fvac()
      -> f = R*Bt in vacuum


Copyright 2016 Ben Dudson, University of York. Email: benjamin.dudson@york.ac.uk

This file is part of FreeGS.

FreeGS is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

FreeGS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with FreeGS.  If not, see <http://www.gnu.org/licenses/>.
"""

from scipy.integrate import romb, quad # Romberg integration
from . import critical
from .gradshafranov import mu0

from numpy import clip, zeros, reshape, sqrt, pi


class Profile(object):
    """
    Base class from which profiles classes can inherit
    
    This provides two methods: 
       pressure(psinorm) and fpol(psinorm)

    which assume that the following methods are available:
       pprime(psinorm), ffprime(psinorm), fvac()
    
    """
    def pressure(self, psinorm, out=None):
        """
        Return p as a function of normalised psi by
        integrating pprime
        """
        
        if not hasattr(psinorm, 'shape'):
            # Assume  a single value
            val, _ = quad(self.pprime, psinorm, 1.0)
            # Convert from integral in normalised psi to integral in psi
            return val * (self.psi_axis - self.psi_bndry)
            
        # Assume a NumPy array
        
        if out is None:
            out = zeros(psinorm.shape)
        
        pvals = reshape(psinorm, -1)
        ovals = reshape(out, -1)

        if len(pvals) != len(ovals):
            raise ValueError("Input and output arrays of different lengths")
        
        for i in range(len(pvals)):
            val, _ = quad(self.pprime, pvals[i], 1.0)
            # Convert from integral in normalised psi to integral in psi
            val *= self.psi_axis - self.psi_bndry
            ovals[i] = val
        
        return reshape(ovals, psinorm.shape)
        
    def fpol(self, psinorm, out=None):
        """
        Return f as a function of normalised psi
        
        """
        
        if not hasattr(psinorm, 'shape'):
            # Assume  a single value
            
            val, _ = quad(self.ffprime, psinorm, 1.0)
            # Convert from integral in normalised psi to integral in psi
            val *= self.psi_axis - self.psi_bndry
            
            # ffprime = 0.5*d/dpsi(f^2)
            # Apply boundary condition at psinorm=1 val = fvac**2
            return sqrt(2.*val + self.fvac()**2)
            
        # Assume it's a NumPy array
        
        if out is None:
            out = zeros(psinorm.shape)
            
        pvals = reshape(psinorm, -1)
        ovals = reshape(out, -1)
        
        if len(pvals) != len(ovals):
            raise ValueError("Input and output arrays of different lengths")
        for i in range(len(pvals)):
            val, _ = quad(self.ffprime, pvals[i], 1.0)
            # Convert from integral in normalised psi to integral in psi
            val *= self.psi_axis - self.psi_bndry
            
            # ffprime = 0.5*d/dpsi(f^2)
            # Apply boundary condition at psinorm=1 val = fvac**2
            ovals[i] = sqrt(2.*val + self.fvac()**2)
            
        return reshape(ovals, psinorm.shape)
    
class ConstrainBetapIp(Profile):
    """
    Constrain poloidal Beta and plasma current

    This is the constraint used in
    YoungMu Jeon arXiv:1503.03135
    
    """

    def __init__(self, betap, Ip, fvac, 
                 alpha_m=1.0, alpha_n=2.0,
                 Raxis=1.0):
        """
        betap - Poloidal beta
        Ip    - Plasma current [Amps]
        fvac  - Vacuum f = R*Bt
        
        Raxis - R used in p' and ff' components
        """
        
        # Check inputs
        if alpha_m < 0:
            raise ValueError("alpha_m must be positive")
        if alpha_n < 0:
            raise ValueError("alpha_n must be positive")

        # Set parameters for later use
        self.betap = betap
        self.Ip = Ip
        self._fvac = fvac
        self.alpha_m = alpha_m
        self.alpha_n = alpha_n
        self.Raxis = Raxis

    def Jtor(self, R, Z, psi):
        """ Calculate toroidal plasma current
        
         Jtor = L * (Beta0*R/Raxis + (1-Beta0)*Raxis/R)*jtorshape
        
         where jtorshape is a shape function
         L and Beta0 are parameters which are set by constraints
        """
        
        # Analyse the equilibrium, finding O- and X-points
        opt, xpt = critical.find_critical(R, Z, psi)
        if not opt:
            raise ValueError("No O-points found!")
        psi_axis = opt[0][2]
        
        if xpt:
            psi_bndry = xpt[0][2]
            mask = critical.core_mask(R, Z, psi, opt, xpt)
        else:
            # No X-points
            psi_bndry = psi[0,0]
            mask = None
        
        dR = R[1,0] - R[0,0]
        dZ = Z[0,1] - Z[0,0]
        
        # Calculate normalised psi.
        # 0 = magnetic axis
        # 1 = plasma boundary
        psi_norm = (psi - psi_axis)  / (psi_bndry - psi_axis)
        
        # Current profile shape
        jtorshape = (1. - psi_norm**self.alpha_m)**self.alpha_n
        
        if mask is not None:
            # If there is a masking function (X-points, limiters)
            jtorshape *= mask
            
        # Now apply constraints to define constants
        
        # Need integral of jtorshape to calculate pressure
        # Note factor to convert from normalised psi integral
        def pshape(psinorm):
            shapeintegral,_ =  quad(lambda x: (1. - x**self.alpha_m)**self.alpha_n, psinorm, 1.0)
            shapeintegral *= (psi_bndry - psi_axis)
            return shapeintegral
        
        # Pressure is
        # 
        # p(psinorm) = - (L*Beta0/Raxis) * pshape(psinorm)
        #
        
        nx,ny = psi_norm.shape
        pfunc = zeros((nx,ny))
        for i in range(1,nx-1):
            for j in range(1,ny-1):
                if (psi_norm[i,j] >= 0.0) and (psi_norm[i,j] < 1.0):
                    pfunc[i,j] = pshape(psi_norm[i,j])
        if mask is not None:
            pfunc *= mask
        
        # Integrate over plasma
        # betap = (8pi/mu0) * int(p)dRdZ / Ip^2
        #       = - (8pi/mu0) * (L*Beta0/Raxis) * intp / Ip^2
        
        intp = romb(romb(pfunc)) * dR*dZ
        
        LBeta0 = -self.betap * (mu0/(8.*pi)) * self.Raxis * self.Ip**2 / intp
        
        # Integrate current components
        IR = romb(romb(jtorshape * R/self.Raxis)) * dR*dZ
        I_R = romb(romb(jtorshape * self.Raxis/R)) * dR*dZ
        
        # Toroidal plasma current Ip is
        #
        # Ip = L * (Beta0 * IR + (1-Beta0)*I_R)
        #    = L*Beta0*(IR - I_R) + L*I_R
        #
        
        L = self.Ip/I_R - LBeta0*(IR/I_R - 1)
        Beta0 = LBeta0 / L

        print("Constraints: L = %e, Beta0 = %e" % (L, Beta0))
        
        # Toroidal current
        Jtor = L * (Beta0*R/self.Raxis + (1-Beta0)*self.Raxis/R)*jtorshape
        
        self.L = L
        self.Beta0 = Beta0
        self.psi_bndry = psi_bndry
        self.psi_axis = psi_axis

        return Jtor

    # Profile functions
    def pprime(self, pn):
        """
        dp/dpsi as a function of normalised psi
        """
        shape = (1. - pn**self.alpha_m)**self.alpha_n
        return self.L * self.Beta0/self.Raxis * shape
        
    def ffprime(self, pn):
        """
        f * df/dpsi as a function of normalised psi
        """
        shape = (1. - pn**self.alpha_m)**self.alpha_n
        return mu0 * self.L * (1-self.Beta0)*self.Raxis * shape
        
        return Jtor, pprime, ffprime
        
    def fvac(self):
        return self._fvac


class ConstrainPaxisIp(Profile):
    """
    Constrain pressure on axis and plasma current
    
    """

    def __init__(self, paxis, Ip, fvac, 
                 alpha_m=1.0, alpha_n=2.0,
                 Raxis=1.0):
        """
        paxis - Pressure at magnetic axis [Pa]
        Ip    - Plasma current [Amps]
        fvac  - Vacuum f = R*Bt
        
        Raxis - R used in p' and ff' components
        """
        
        # Check inputs
        if alpha_m < 0:
            raise ValueError("alpha_m must be positive")
        if alpha_n < 0:
            raise ValueError("alpha_n must be positive")

        # Set parameters for later use
        self.paxis = paxis
        self.Ip = Ip
        self._fvac = fvac
        self.alpha_m = alpha_m
        self.alpha_n = alpha_n
        self.Raxis = Raxis

    def Jtor(self, R, Z, psi):
        """ Calculate toroidal plasma current
        
         Jtor = L * (Beta0*R/Raxis + (1-Beta0)*Raxis/R)*jtorshape
        
         where jtorshape is a shape function
         L and Beta0 are parameters which are set by constraints
        """
        
        # Analyse the equilibrium, finding O- and X-points
        opt, xpt = critical.find_critical(R, Z, psi)
        if not opt:
            raise ValueError("No O-points found!")
        psi_axis = opt[0][2]
        
        if xpt:
            psi_bndry = xpt[0][2]
            mask = critical.core_mask(R, Z, psi, opt, xpt)
        else:
            # No X-points
            psi_bndry = psi[0,0]
            mask = None
        
        dR = R[1,0] - R[0,0]
        dZ = Z[0,1] - Z[0,0]
        
        # Calculate normalised psi.
        # 0 = magnetic axis
        # 1 = plasma boundary
        psi_norm = (psi - psi_axis)  / (psi_bndry - psi_axis)
        
        # Current profile shape
        jtorshape = (1. - psi_norm**self.alpha_m)**self.alpha_n
        
        if mask is not None:
            # If there is a masking function (X-points, limiters)
            jtorshape *= mask
            
        # Now apply constraints to define constants
        
        # Need integral of jtorshape to calculate paxis
        # Note factor to convert from normalised psi integral
        shapeintegral,_ =  quad(lambda x: (1. - x**self.alpha_m)**self.alpha_n, 0.0, 1.0)
        shapeintegral *= (psi_bndry - psi_axis)
        
        # Pressure on axis is
        # 
        # paxis = - (L*Beta0/Raxis) * shapeintegral
        #

        # Integrate current components
        IR = romb(romb(jtorshape * R/self.Raxis)) * dR*dZ
        I_R = romb(romb(jtorshape * self.Raxis/R)) * dR*dZ
        
        # Toroidal plasma current Ip is
        #
        # Ip = L * (Beta0 * IR + (1-Beta0)*I_R)
        #    = L*Beta0*(IR - I_R) + L*I_R
        #
        
        LBeta0 = -self.paxis*self.Raxis / shapeintegral
        
        L = self.Ip/I_R - LBeta0*(IR/I_R - 1)
        Beta0 = LBeta0 / L

        print("Constraints: L = %e, Beta0 = %e" % (L, Beta0))
        
        # Toroidal current
        Jtor = L * (Beta0*R/self.Raxis + (1-Beta0)*self.Raxis/R)*jtorshape
        
        self.L = L
        self.Beta0 = Beta0
        self.psi_bndry = psi_bndry
        self.psi_axis = psi_axis

        return Jtor

    # Profile functions
    def pprime(self, pn):
        """
        dp/dpsi as a function of normalised psi
        """
        shape = (1. - pn**self.alpha_m)**self.alpha_n
        return self.L * self.Beta0/self.Raxis * shape
        
    def ffprime(self, pn):
        """
        f * df/dpsi as a function of normalised psi
        """
        shape = (1. - pn**self.alpha_m)**self.alpha_n
        return mu0 * self.L * (1-self.Beta0)*self.Raxis * shape
        
        return Jtor, pprime, ffprime
        
    def fvac(self):
        return self._fvac
        



class ProfilesPprimeFfprime:
    """
    Specified profile functions p'(psi), ff'(psi)
    
    Jtor = R*p' + ff'/(R*mu0)
    
    """
    def __init__(self, pprime_func, ffprime_func, fvac, p_func=None, f_func=None):
        """
        pprime_func(psi_norm) - A function which returns dp/dpsi at given normalised flux
        ffprime_func(psi_norm) - A function which returns f*df/dpsi at given normalised flux (f = R*Bt)

        fvac - Vacuum f = R*Bt
        
        Optionally, the pres
        """
        self.pprime = pprime_func
        self.ffprime = ffprime_func
        self.p_func = p_func
        self.f_func = f_func
        self._fvac = fvac
        
    def Jtor(self, R, Z, psi):
        """
        Calculate toroidal plasma current
        
        Jtor = R*p' + ff'/(R*mu0)
        """
        
        # Analyse the equilibrium, finding O- and X-points
        opt, xpt = critical.find_critical(R, Z, psi)
        if not opt:
            raise ValueError("No O-points found!")
        psi_axis = opt[0][2]
        
        if xpt:
            psi_bndry = xpt[0][2]
            mask = critical.core_mask(R, Z, psi, opt, xpt)
        else:
            # No X-points
            psi_bndry = psi[0,0]
            mask = None
        
        dR = R[1,0] - R[0,0]
        dZ = Z[0,1] - Z[0,0]
        
        # Calculate normalised psi.
        # 0 = magnetic axis
        # 1 = plasma boundary
        psi_norm = clip((psi - psi_axis)  / (psi_bndry - psi_axis), 0.0, 1.0)
        Jtor = R * self.pprime(psi_norm) + self.ffprime(psi_norm)/(R * mu0)

        if mask is not None:
            # If there is a masking function (X-points, limiters)
            Jtor *= mask
        
        return Jtor

    def pressure(self, psinorm, out=None):
        """
        Return pressure [Pa] at given value(s) of
        normalised psi.
        """
        if self.p_func is not None:
            # If a function exists then use it
            return self.p_func(psinorm)
        
        # If not, use base class to integrate
        return super(ProfilesPprimeFfprime, self).pressure(psinorm, out)
        
    def fpol(self, psinorm, out=None):
        """
        Return f=R*Bt at given value(s) of
        normalised psi.
        """
        if self.f_func is not None:
            # If a function exists then use it
            return self.f_func(psinorm)
        
        # If not, use base class to integrate
        return super(ProfilesPprimeFfprime, self).fpol(psinorm, out)
    def fvac(self):
        return self._fvac

        



"""
Defines class to represent the equilibrium
state, including plasma and coil currents
"""

from numpy import meshgrid, linspace, exp, zeros, nditer
from scipy import interpolate
from scipy.integrate import romb, quad # Romberg integration

from .boundary import fixedBoundary, freeBoundary

# Operators which define the G-S equation
from .gradshafranov import mu0, GSsparse

# Multigrid solver
from . import multigrid

from . import machine

class Equilibrium:
    """
    Represents the equilibrium state, including
    plasma and coil currents
    
    Data members
    ------------
    
    These can be read, but should not be modified directly
    
    R[nx,ny]
    Z[nx,ny] 
    
    Rmin, Rmax
    Zmin, Zmax

    tokamak - The coils and circuits
    
    Private data members
    
    _applyBoundary()
    _solver - Grad-Shafranov elliptic solver
    
    """
    
    def __init__(self, tokamak=machine.EmptyTokamak(), 
                 Rmin=0.1, Rmax=2.0,
                 Zmin=-1.0, Zmax=1.0,
                 nx=65, ny=65,
                 boundary=freeBoundary):
        """
        Initialises a plasma equilibrium
        
        Rmin, Rmax  - Range of major radius R [m]
        Zmin, Zmax  - Range of height Z [m]
        
        nx - Resolution in R. This must be 2^n + 1
        ny - Resolution in Z. This must be 2^m + 1
        
        boundary - The boundary condition, either freeBoundary or fixedBoundary
        
        """

        self.tokamak = tokamak

        self._applyBoundary = boundary
        
        self.Rmin = Rmin
        self.Rmax = Rmax
        self.Zmin = Zmin
        self.Zmax = Zmax
        
        dR = (Rmax - Rmin)/(nx - 1)
        dZ = (Zmax - Zmin)/(ny - 1)
        
        yymid = 1 - Zmax/(Zmax - Zmin)
        
        # Starting guess for psi
        xx, yy = meshgrid(linspace(0,1,nx), linspace(0,1,ny), indexing='ij')
        psi = exp( - ( (xx - 0.5)**2 + (yy - yymid)**2 ) / 0.4**2 )
        
        psi[0,:] = 0.0
        psi[:,0] = 0.0
        psi[-1,:] = 0.0
        psi[:,-1] = 0.0
        
        self.R = zeros([nx,ny])
        for x in range(nx):
            self.R[x,:] = Rmin + x * dR
            
        self.Z = zeros([nx,ny])
        for y in range(ny):
            self.Z[:,y] = Zmin + y*dZ
            
        self._updatePlasmaPsi(psi)
        
        # Calculate coil Greens functions. This is an optimisation,
        # used in self.psi() to speed up calculations
        self._pgreen = tokamak.createPsiGreens(self.R, self.Z)

        self._current = 0.0 # Plasma current

        # Create the solver
        generator = GSsparse(Rmin, Rmax, Zmin, Zmax)
        self._solver = multigrid.createVcycle(nx, ny,
                                              generator,
                                              nlevels=1,
                                              ncycle=1,
                                              niter=2,
                                              direct=True)
        
    def setSolverVcycle(nlevels=1, ncycle=1, niter=1, direct=True):
        """
        Creates a new linear solver, based on the multigrid code
    
        nlevels  - Number of resolution levels, including original
        ncycle   - The number of V cycles
        niter    - Number of linear solver (Jacobi) iterations per level
        direct   - Use a direct solver at the coarsest level?
        
        """
        generator = GSsparse(Rmin, Rmax, Zmin, Zmax)
        nx,ny = self.R.shape
        
        self._solver = multigrid.createVcycle(nx, ny,
                                              generator,
                                              nlevels=nlevels,
                                              ncycle=ncycle,
                                              niter=niter,
                                              direct=direct)
        
    def setSolver(solver):
        """
        Sets the linear solver to use. The given object/function must have a __call__ method
        which takes two inputs

        solver(x, b)
        
        where x is the initial guess. This should solve Ax = b, returning the result.

        """
        self._solver = solver

    def callSolver(self, psi, rhs):
        """
        Calls the psi solver, passing the initial guess and RHS arrays
        
        psi   Initial guess for the solution (used if iterative)
        rhs   
        
        Returns
        -------
        
        Solution psi

        """
        return self._solver(psi, rhs)
                        
    def getMachine(self):
        """
        Returns the handle of the machine, including coils
        """
        return self.tokamak
    
    def plasmaCurrent(self):
        """
        Plasma current [Amps]
        """
        return self._current
    
    def plasmaBr(self, R,Z):
        """
        Radial magnetic field due to plasma
        Br = -1/R dpsi/dZ
        """
        return -self.psi_func(R,Z,dy=1)[0][0]/R
        
    def plasmaBz(self, R, Z):
        """
        Vertical magnetic field due to plasma 
        Bz = (1/R) dpsi/dR
        """
        return self.psi_func(R,Z,dx=1)[0][0]/R
        
    def Br(self, R, Z):
        """
        Total radial magnetic field
        """
        return self.plasmaBr(R,Z) + self.tokamak.Br(R,Z)

    def Bz(self, R, Z):
        """
        Total vertical magnetic field
        """
        return self.plasmaBz(R,Z) + self.tokamak.Bz(R,Z)
    
    def psi(self):
        #return self.plasma_psi + self.tokamak.psi(self.R, self.Z)
        return self.plasma_psi + self.tokamak.calcPsiFromGreens(self._pgreen)
        
    def psiRZ(self, R, Z):
        """
        Return poloidal flux psi at given (R,Z) location
        """
        return self.psi_func(R,Z) + self.tokamak.psi(R,Z)

    def fpol(self,psinorm):
        """
        Return f = R*Bt at specified values of normalised psi
        """
        return self._profiles.fpol(psinorm)
    
    def fvac(self):
        """
        Return vacuum f = R*Bt
        """
        return self._profiles.fvac()
    
    def q(self, psinorm):
        """
        Returns safety factor q at specified values of normalised psi
        """
        return psinorm * 0.0
        
    def pprime(self, psinorm):
        """
        Return p' at given normalised psi
        """
        return self._profiles.pprime(psinorm)
    
    def ffprime(self, psinorm):
        """
        Return ff' at given normalised psi
        """
        return self._profiles.ffprime(psinorm)
        
    def pressure(self, psinorm, out=None):
        """
        Returns plasma pressure at specified values of normalised psi
        """
        return self._profiles.pressure(psinorm)


    def solve(self, profiles):
        """
        Calculate the plasma equilibrium given new profiles
        replacing the current equilibrium.
        
        This performs the linear Grad-Shafranov solve
        
        profiles  - An object describing the plasma profiles.
                    At minimum this must have methods:
             .Jtor(R, Z, psi)   -> [nx, ny]
             .pprime(psinorm)
             .ffprime(psinorm)
             .pressure(psinorm)
             .fpol(psinorm)
        """
        
        self._profiles = profiles
        
        # Calculate toroidal current density
        Jtor = profiles.Jtor(self.R, self.Z, self.psi())
        
        # Set plasma boundary
        # Note that the Equilibrium is passed to the boundary function
        # since the boundary may need to run the G-S solver (von Hagenow's method)
        self._applyBoundary(self, Jtor, self.plasma_psi)
        
        # Right hand side of G-S equation
        rhs = -mu0 * self.R * Jtor

        # Copy boundary conditions
        rhs[0,:] = self.plasma_psi[0,:]
        rhs[:,0] = self.plasma_psi[:,0]
        rhs[-1,:] = self.plasma_psi[-1,:]
        rhs[:,-1] = self.plasma_psi[:,-1]
        
        # Call elliptic solver
        plasma_psi = self._solver(self.plasma_psi, rhs)
        
        self._updatePlasmaPsi(plasma_psi)

        # Update plasma current
        dR = self.R[1,0] - self.R[0,0]
        dZ = self.Z[0,1] - self.Z[0,0]
        self._current = romb(romb(Jtor)) * dR*dZ
        
    def _updatePlasmaPsi(self, plasma_psi):
        """
        Sets the plasma psi data, updates spline interpolation coefficients
        """
        self.plasma_psi = plasma_psi

        # Update spline interpolation
        self.psi_func = interpolate.RectBivariateSpline(self.R[:,0], self.Z[0,:], plasma_psi)
     


def refine(eq):
    """
    Double grid resolution, returning a new equilibrium
    
    """
    # Interpolate the plasma psi
    plasma_psi = multigrid.interpolate(eq.plasma_psi)
    nx, ny = plasma_psi.shape
    
    result = Equilibrium(tokamak=eq.tokamak,
                         Rmin = eq.Rmin,
                         Rmax = eq.Rmax,
                         Zmin = eq.Zmin,
                         Zmax = eq.Zmax,
                         nx=nx, ny=ny)

    result._updatePlasmaPsi(plasma_psi)
    
    if hasattr(eq, "_profiles"):
        result._profiles = eq._profiles

    if hasattr(eq, "control"):
        result.control = eq.control

    return result

def coarsen(eq):
    """
    Reduce grid resolution, returning a new equilibrium
    """
    plasma_psi = multigrid.restrict(eq.plasma_psi)
    nx, ny = plasma_psi.shape
    
    result = Equilibrium(tokamak=eq.tokamak,
                         Rmin = eq.Rmin,
                         Rmax = eq.Rmax,
                         Zmin = eq.Zmin,
                         Zmax = eq.Zmax,
                         nx=nx, ny=ny)

    result._updatePlasmaPsi(plasma_psi)
    
    if hasattr(eq, "_profiles"):
        result._profiles = eq._profiles

    if hasattr(eq, "control"):
        result.control = eq.control

    return result

if __name__=="__main__":
    
    # Test the different spline interpolation routines
    
    from numpy import ravel
    import matplotlib.pyplot as plt
    
    import machine
    tokamak = machine.TestTokamak()

    Rmin=0.1
    Rmax=2.0

    eq = Equilibrium(tokamak, Rmin=Rmin, Rmax=Rmax)
    
    import constraints
    xpoints = [(1.2, -0.8),
               (1.2, 0.8)]
    constraints.xpointConstrain(eq, xpoints)
    
    psi = eq.psi()

    tck = interpolate.bisplrep(ravel(eq.R), ravel(eq.Z), ravel(psi))
    spline = interpolate.RectBivariateSpline(eq.R[:,0], eq.Z[0,:], psi)
    f = interpolate.interp2d(eq.R[:,0], eq.Z[0,:],psi, kind='cubic')

    plt.plot(eq.R[:,10], psi[:,10], 'o')

    r = linspace(Rmin, Rmax, 1000)
    z = eq.Z[0,10]
    plt.plot(r, f(r,z), label="f")
    
    plt.plot(r, spline(r,z), label="spline")
    
    plt.plot(r, interpolate.bisplev(r,z, tck), label="bisplev")
    
    plt.legend()
    plt.show()



"""
FreeGS
======

Free boundary Grad-Shafranov solver


.. moduleauthor:: Ben Dudson <benjamin.dudson@york.ac.uk>


License
-------

Copyright 2016 Ben Dudson, University of York. Email: benjamin.dudson@york.ac.uk

FreeGS is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

FreeGS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with FreeGS.  If not, see <http://www.gnu.org/licenses/>.

"""

__version__ = "0.1.0"

from .equilibrium import Equilibrium

from . import jtor

from . import machine

from . import control

from .picard import solve




"""
Plotting using matplotlib

Copyright 2016 Ben Dudson, University of York. Email: benjamin.dudson@york.ac.uk

This file is part of FreeGS.

FreeGS is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

FreeGS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with FreeGS.  If not, see <http://www.gnu.org/licenses/>.

"""

import matplotlib.pyplot as plt

from numpy import linspace, amin, amax
from . import critical

def plotEquilibrium(eq, axis=None, show=True, oxpoints=True):
    """
    Plot the equilibrium flux surfaces

    axis - Specify the axis on which to plot
    
    """

    R = eq.R
    Z = eq.Z
    psi = eq.psi()

    if axis is None:
        fig = plt.figure()
        axis = fig.add_subplot(111)

    levels = linspace(amin(psi), amax(psi), 100)
    
    axis.contour(R,Z,psi, levels=levels)
    axis.set_aspect('equal')
    axis.set_xlabel("Major radius [m]")
    axis.set_ylabel("Height [m]")

    if oxpoints:
        # Add O- and X-points
        opt, xpt = critical.find_critical(eq.R, eq.Z, psi)
        
        for r,z,_ in xpt:
            axis.plot(r,z,'ro')
        for r,z,_ in opt:
            axis.plot(r,z,'go')
            
        if xpt:
            psi_bndry = xpt[0][2]
            sep_contour=axis.contour(eq.R, eq.Z,psi, levels=[psi_bndry], colors='r')
        
    if show:
        plt.show()
    
    return axis

def plotCoils(coils, axis=None):
    if axis is None:
        fig = plt.figure()
        axis = fig.add_subplot(111)
     
        
    return axis


def plotConstraints(xpoints, axis=None):
    """
    Plots constraints used for coil currents
    
    """

    if axis is None:
        fig = plt.figure()
        axis = fig.add_subplot(111)

    for xpt in xpoints:
        axis.plot(xpt[0], xpt[1], 'ro')    

    return axis



"""
Grad-Shafranov equation

Copyright 2016 Ben Dudson, University of York. Email: benjamin.dudson@york.ac.uk

This file is part of FreeGS.

FreeGS is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

FreeGS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with FreeGS.  If not, see <http://www.gnu.org/licenses/>.

"""

from numpy import zeros

# Elliptic integrals of first and second kind (K and E)
from scipy.special import ellipk, ellipe

from numpy import sqrt, pi, clip

from scipy.sparse import lil_matrix, eye

# Physical constants
mu0 = 4e-7*pi

class GSElliptic:
    """
    Represents the Grad-Shafranov elliptic operator
    
    \Delta^* = R^2 \nabla\cdot\frac{1}{R^2}\nabla
    
    """

    def __init__(self, Rmin):
        """
        Specify minimum radius
        
        """
        self.Rmin = Rmin
        
    def __call__(self, psi, dR, dZ):
        """
        Apply the full operator to 2D field f
        
        Inputs
        ------
        
        psi[R,Z]   A 2D NumPy array containing flux function psi
        dR         A scalar (uniform grid spacing)
        dZ         A scalar (uniform grid spacing)
        
        """
        nx = psi.shape[0]
        ny = psi.shape[1]
        
        b = zeros([nx,ny])

        invdR2 = 1./dR**2
        invdZ2 = 1./dZ**2
        
        for x in range(1,nx-1):
            R = self.Rmin + dR*x  # Major radius of this point
            for y in range(1,ny-1):
                # Loop over points in the domain
                b[x,y] = ( psi[x,y-1]*invdZ2
                           + (invdR2 + 1./(2.*R*dR))*psi[x-1,y]
                           - 2.*(invdR2 + invdZ2) * psi[x,y]
                           + (invdR2 - 1./(2.*R*dR))*psi[x+1,y]
                           + psi[x,y+1]*invdZ2 )
        return b
        
    def diag(self, dR, dZ):
        """
        Return the diagonal elements
        
        """
        return -2./dR**2 - 2./dZ**2

class GSsparse:
    """
    Calculates sparse matrices for the Grad-Shafranov operator
    """
    def __init__(self, Rmin, Rmax, Zmin, Zmax):
        self.Rmin = Rmin
        self.Rmax = Rmax
        self.Zmin = Zmin
        self.Zmax = Zmax

    def __call__(self, nx, ny):
        """
        Create a sparse matrix with given resolution
        
        """

        # Calculate grid spacing
        dR = (self.Rmax - self.Rmin)/(nx - 1)
        dZ = (self.Zmax - self.Zmin)/(ny - 1)
        
        # Total number of points
        N = nx * ny

        # Create a linked list sparse matrix
        A = eye(N, format="lil")
        
        invdR2 = 1./dR**2
        invdZ2 = 1./dZ**2
        
        for x in range(1,nx-1):
            R = self.Rmin + dR*x  # Major radius of this point
            for y in range(1,ny-1):
                # Loop over points in the domain
                row = x*ny + y

                # y-1 
                A[row, row-1] = invdZ2

                # x-1
                A[row, row-ny] = (invdR2 + 1./(2.*R*dR))

                # diagonal
                A[row, row] = - 2.*(invdR2 + invdZ2)

                # x+1
                A[row, row+ny] = (invdR2 - 1./(2.*R*dR))

                # y+1
                A[row, row+1] = invdZ2
                
        # Convert to Compressed Sparse Row (CSR) format
        return A.tocsr()
    
def Greens(Rc, Zc, R, Z):
    """
    Calculate poloidal flux at (R,Z) due to a unit current
    at (Rc,Zc) using Greens function
    
    """

    # Calculate k^2
    k2 = 4.*R * Rc / ( (R + Rc)**2 + (Z - Zc)**2 )

    # Clip to between 0 and 1 to avoid nans e.g. when coil is on grid point
    k2 = clip(k2, 1e-10, 1.0 - 1e-10)
    k = sqrt(k2)

    # Note definition of ellipk, ellipe in scipy is K(k^2), E(k^2)
    return (mu0/(2.*pi)) * sqrt(R*Rc) * ( (2. - k2)*ellipk(k2) - 2.*ellipe(k2) ) / k

def GreensBz(Rc, Zc, R, Z, eps=1e-3):
    """
    Calculate radial magnetic field at (R,Z)
    due to unit current at (Rc, Zc)
    
    Bz = (1/R) d psi/dR
    """
    
    return (Greens(Rc, Zc, R+eps, Z) - Greens(Rc, Zc, R-eps, Z))/(2.*eps*R)

def GreensBr(Rc, Zc, R, Z, eps=1e-3):
    """
    Calculate vertical magnetic field at (R,Z)
    due to unit current at (Rc, Zc)
    
    Br = -(1/R) d psi/dZ
    """
    
    return (Greens(Rc, Zc, R, Z-eps) - Greens(Rc, Zc, R, Z+eps))/(2.*eps*R)
